package org.xbib.event.async.yield.primitives.intgr;

import org.xbib.event.async.yield.Traverser;
import org.xbib.event.async.yield.Yield;
import org.xbib.event.async.yield.primitives.dbl.DoubleTraverser;
import org.xbib.event.async.yield.primitives.lng.LongTraverser;

import java.util.function.DoubleToIntFunction;
import java.util.function.LongToIntFunction;
import java.util.function.ToIntFunction;

/**
 * Bulk traversal.
 * We use a traverse method as its first choice to
 * implement Query operations.
 * This is a special kind of traversal that disallows individually access.
 */
public interface IntTraverser extends Traverser<Integer> {
    static IntTraverser empty() {
        return yield -> {
        };
    }

    /**
     * An IntTraverser object from a generic {@link Traverser} mapped by a {@link ToIntFunction}.
     *
     * @param source {@link Traverser} with the source elements for this {@code IntTraverser}.
     * @param mapper {@link ToIntFunction} that specifies how to map the source elements into int values.
     */
    static <T> IntTraverser from(Traverser<T> source, ToIntFunction<? super T> mapper) {
        return yield -> source.traverse(item -> yield.ret(mapper.applyAsInt(item)));
    }

    /**
     * An IntTraverser object from a {@link DoubleTraverser} mapped by a {@link DoubleToIntFunction}.
     *
     * @param source {@link DoubleTraverser} with the source elements for this {@code IntTraverser}.
     * @param mapper {@link DoubleToIntFunction} that specifies how to map the source elements into int values.
     */
    static IntTraverser from(DoubleTraverser source, DoubleToIntFunction mapper) {
        return from((Traverser<Double>) source, mapper::applyAsInt);
    }

    /**
     * An IntTraverser object from a {@link LongTraverser} mapped by a {@link LongToIntFunction}.
     *
     * @param source {@link LongTraverser} with the source elements for this {@code DoubleTraverser}.
     * @param mapper {@link LongToIntFunction} that specifies how to map the source elements into int values.
     */
    static IntTraverser from(LongTraverser source, LongToIntFunction mapper) {
        return from((Traverser<Long>) source, mapper::applyAsInt);
    }

    /**
     * Yields elements sequentially in the current thread,
     * until all elements have been processed or an
     * exception is thrown.
     */
    void traverse(IntYield yield);

    /**
     * Default traverse implementation that calls the
     * primitive version of it
     */
    @Override
    default void traverse(Yield<? super Integer> yield) {
        IntYield yld = yield::ret;
        this.traverse(yld);
    }
}
