package org.xbib.event.async.yield.primitives.lng;

import org.xbib.event.async.yield.Traverser;
import org.xbib.event.async.yield.Yield;
import org.xbib.event.async.yield.primitives.dbl.DoubleTraverser;
import org.xbib.event.async.yield.primitives.intgr.IntAdvancer;
import org.xbib.event.async.yield.primitives.intgr.IntTraverser;

import java.util.function.DoubleToLongFunction;
import java.util.function.IntToLongFunction;
import java.util.function.ToLongFunction;

/**
 * Bulk traversal.
 * We use traverse method as its first choice to
 * implement Query operations.
 * This is a special kind of traversal that disallows individually access.
 */
public interface LongTraverser extends Traverser<Long> {
    /**
     * An LongTraverser object without elements.
     */
    static LongTraverser empty() {
        return yield -> {
        };
    }

    /**
     * An LongTraverser object from a generic {@link Traverser} mapped by a {@link ToLongFunction}.
     *
     * @param source {@link Traverser} with the source elements for this {@code LongTraverser}.
     * @param mapper {@link ToLongFunction} that specifies how to map the source elements longo long values.
     */
    static <T> LongTraverser from(Traverser<T> source, ToLongFunction<? super T> mapper) {
        return yield -> source.traverse(item -> yield.ret(mapper.applyAsLong(item)));
    }

    /**
     * An LongTraverser object from a {@link DoubleTraverser} mapped by a {@link DoubleToLongFunction}.
     *
     * @param source {@link DoubleTraverser} with the source elements for this {@code LongAdvancer}.
     * @param mapper {@link DoubleToLongFunction} that specifies how to map the source elements longo long values.
     */
    static LongTraverser from(DoubleTraverser source, DoubleToLongFunction mapper) {
        return from((Traverser<Double>) source, mapper::applyAsLong);
    }

    /**
     * An LongAdvancer object from a {@link IntTraverser} mapped by a {@link IntToLongFunction}.
     *
     * @param source {@link IntAdvancer} with the source elements for this {@code LongAdvancer}.
     * @param mapper {@link IntToLongFunction} that specifies how to map the source elements longo long values.
     */
    static LongTraverser from(IntTraverser source, IntToLongFunction mapper) {
        return from((Traverser<Integer>) source, mapper::applyAsLong);
    }

    /**
     * Default traverse implementation that calls the
     * primitive version of it
     */
    @Override
    default void traverse(Yield<? super Long> yield) {
        LongYield yld = yield::ret;
        this.traverse(yld);
    }

    /**
     * Yields elements sequentially in the current thread,
     * until all elements have been processed or an
     * exception is thrown.
     */
    void traverse(LongYield yield);
}
