package org.xbib.event.log;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.text.MessageFormat;
import java.time.Instant;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.zip.GZIPOutputStream;

public class RotatingFileOutputStream extends OutputStream implements Rotatable {

    private static final Logger logger = Logger.getLogger(RotatingFileOutputStream.class.getName());

    private final RotationConfig config;

    private final List<RotationPolicy> writeSensitivePolicies;

    private final ScheduledThreadPoolExecutor scheduledThreadPoolExecutor;

    private volatile ByteCountingOutputStream byteCountingOutputStream;

    public RotatingFileOutputStream(RotationConfig config) throws IOException {
        this(config, createDefaultExecutorService());
    }

    public RotatingFileOutputStream(RotationConfig config, ScheduledThreadPoolExecutor scheduledThreadPoolExecutor) throws IOException {
        this.config = config;
        this.scheduledThreadPoolExecutor = scheduledThreadPoolExecutor;
        this.writeSensitivePolicies = config.getPolicies().stream().filter(RotationPolicy::isWriteSensitive).collect(Collectors.toList());
        this.byteCountingOutputStream = open(null, config.getClock().now());
        startPolicies();
    }

    @Override
    public void rotate(RotationPolicy policy, Instant instant) throws IOException {
        Objects.requireNonNull(instant, "instant");
        RotationListener listener = config.getListener();
        listener.onTrigger(policy, instant);
        byteCountingOutputStream.flush();
        if (Files.size(config.getPath()) == 0) {
            logger.log(Level.FINE, MessageFormat.format("empty file, skipping rotation {path={}}", config.getPath()));
            return;
        }
        listener.onClose(policy, instant);
        byteCountingOutputStream.close();
        Path rotated = config.getFilePattern().create(instant);
        logger.log(Level.FINE, MessageFormat.format("renaming {file={0}, rotatedFile={1}}", config.getPath(), rotated));
        Files.move(config.getPath(), rotated);
        logger.log(Level.FINE, MessageFormat.format("re-opening file {path={0}}", config.getPath()));
        byteCountingOutputStream = open(policy, instant);
        if (config.isCompress()) {
            compress(policy, instant, rotated, listener);
            return;
        }
        listener.onSuccess(policy, instant, rotated);
    }

    @Override
    public void onException(RotationPolicy policy, Instant instant, Path path, Throwable throwable) {
        config.getListener().onFailure(policy, instant, path, throwable);
    }

    private void compress(RotationPolicy policy, Instant instant, Path rotated, RotationListener listener) {
        scheduledThreadPoolExecutor.execute(() -> {
            Path compressed = Paths.get(rotated + ".gz");
            logger.log(Level.FINE, MessageFormat.format("compressing {rotated={}, compressed={}}", rotated, compressed));
            try (InputStream inputStream = Files.newInputStream(rotated);
                 GZIPOutputStream gzipOutputStream = new GZIPOutputStream(Files.newOutputStream(compressed))) {
                inputStream.transferTo(gzipOutputStream);
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                listener.onFailure(policy, instant, rotated, e);
            }
            logger.log(Level.FINE, MessageFormat.format("deleting old file {rotated={}}", rotated));
            try {
                Files.delete(rotated);
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                listener.onFailure(policy, instant, rotated, e);
            }
            listener.onSuccess(policy, instant, compressed);
        });
    }

    @Override
    public RotationConfig getConfig() {
        return config;
    }

    @Override
    public synchronized void write(int b) throws IOException {
        long byteCount = byteCountingOutputStream.size() + 1;
        notifyWriteSensitivePolicies(byteCount);
        byteCountingOutputStream.write(b);
    }

    @Override
    public synchronized void write(byte[] b) throws IOException {
        long byteCount = byteCountingOutputStream.size() + b.length;
        notifyWriteSensitivePolicies(byteCount);
        byteCountingOutputStream.write(b);
    }

    @Override
    public synchronized void write(byte[] b, int off, int len) throws IOException {
        long byteCount = byteCountingOutputStream.size() + len;
        notifyWriteSensitivePolicies(byteCount);
        byteCountingOutputStream.write(b, off, len);
    }

    @Override
    public synchronized void flush() throws IOException {
        byteCountingOutputStream.flush();
    }

    @Override
    public synchronized void close() throws IOException {
        config.getListener().onClose(null, config.getClock().now());
        if (byteCountingOutputStream != null) {
            byteCountingOutputStream.close();
            byteCountingOutputStream = null;
        }
    }

    private static ScheduledThreadPoolExecutor createDefaultExecutorService() {
        final AtomicInteger threadCount = new AtomicInteger();
        return new ScheduledThreadPoolExecutor(4, runnable -> {
            Thread thread = new Thread(runnable, "file-rotation-" + threadCount.getAndIncrement());
            thread.setDaemon(true);
            return thread;
        });
    }

    private void startPolicies() {
        for (RotationPolicy policy : config.getPolicies()) {
            policy.start(this);
        }
    }

    private ByteCountingOutputStream open(RotationPolicy policy, Instant instant) throws IOException {
        OutputStream fileOutputStream = config.isAppend() && Files.exists(config.getPath()) ?
                Files.newOutputStream(config.getPath(), StandardOpenOption.APPEND) :
                Files.newOutputStream(config.getPath());
        config.getListener().onOpen(policy, instant);
        long size = config.isAppend() ? Files.size(config.getPath()) : 0;
        return new ByteCountingOutputStream(fileOutputStream, size);
    }

    private void notifyWriteSensitivePolicies(long byteCount) {
        for (RotationPolicy writeSensitivePolicy : writeSensitivePolicies) {
            writeSensitivePolicy.acceptWrite(byteCount);
        }
    }

    @Override
    public String toString() {
        return "RotatingFileOutputStream " + config.getPath();
    }
}
