package org.xbib.event.persistence;

import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public class AbstractPersistenceStore implements PersistenceStore<String, Object> {

    private final Map<String, Object> internalMap;

    private final Durability durability;

    private final Codec<Map<String, Object>> codec;

    private final String storeName;

    public AbstractPersistenceStore(Codec<Map<String, Object>> codec,
                                    Durability durability,
                                    String storeName) {
        super();
        this.internalMap = new LinkedHashMap<>();
        this.codec = codec;
        this.durability = durability;
        this.storeName = storeName;
    }

    @Override
    public Durability getDurability() {
        return durability;
    }

    @Override
    public Codec<Map<String, Object>> getCodec() {
        return codec;
    }

    @Override
    public void load() throws IOException {
        internalMap.clear();
        Map<String, Object> map = codec.read(storeName);
        if (map != null) {
            internalMap.putAll(map);
        }
    }

    @Override
    public void begin() {
        // ignore
    }

    @Override
    public void commit() throws IOException {
        codec.write(storeName, internalMap);
    }

    @Override
    public void rollback() {
        // ignore
    }

    @SuppressWarnings("unchecked")
    @Override
    public void insert(String key, Object value) throws IOException {
        internalMap.putIfAbsent(key, new ArrayList<>());
        List<Object> list = (List<Object>) internalMap.get(key);
        list.add(value);
        internalMap.put(key, list);
        commit();
    }

    @SuppressWarnings("unchecked")
    @Override
    public boolean remove(String key, Object value) throws IOException {
        internalMap.putIfAbsent(key, new ArrayList<>());
        List<Object> list = (List<Object>) internalMap.get(key);
        boolean b = list.remove(value);
        internalMap.put(key, list);
        commit();
        return b;
    }

    @Override
    public Object get(String key) {
        return internalMap.get(key);
    }

    @Override
    public Object getOrDefault(String key, Object defaultValue) {
        return internalMap.getOrDefault(key, defaultValue);
    }

    @Override
    public void clear() {
        internalMap.clear();
    }
}
