package org.xbib.event.syslog;

import java.time.LocalDateTime;
import java.util.regex.Matcher;

public final class RFC3164MessageParser extends MessageParser {

    private static final String PATTERN = "^(<(?<priority>\\d+)>)?(?<date>([a-zA-Z]{3}\\s+\\d+\\s+\\d+:\\d+:\\d+)|([0-9T:.Z-]+))\\s+(?<host>\\S+)\\s+((?<tag>[^\\[\\s\\]]+)(\\[(?<procid>\\d+)\\])?:)*\\s*(?<message>.+)$";

    private final ThreadLocal<Matcher> matcherThreadLocal;

    public RFC3164MessageParser() {
        this.matcherThreadLocal = initMatcher(PATTERN);
    }

    @Override
    public Message parse(SyslogRequest request) {
        final Matcher matcher = matcherThreadLocal.get().reset(request.rawMessage());
        if (!matcher.find()) {
            return null;
        }
        final String groupPriority = matcher.group("priority");
        final String groupDate = matcher.group("date");
        final String groupHost = matcher.group("host");
        final String groupMessage = matcher.group("message");
        final String groupTag = matcher.group("tag");
        final String groupProcId = matcher.group("procid");
        final String processId = (groupProcId == null || groupProcId.isEmpty()) ? null : groupProcId;
        final Integer priority = (groupPriority == null || groupPriority.isEmpty()) ? null : Integer.parseInt(groupPriority);
        final Integer facility = null == priority ? null : Priority.facility(priority);
        final Integer level = null == priority ? null : Priority.level(priority, facility);
        final LocalDateTime date = parseDate(groupDate);
        return DefaultSyslogMessage.builder()
                .type(MessageType.RFC3164)
                .rawMessage(request.rawMessage())
                .remoteAddress(request.remoteAddress())
                .date(date)
                .host(groupHost)
                .level(level)
                .facility(facility)
                .message(groupMessage)
                .tag(groupTag)
                .processId(processId)
                .build();
    }
}
